import json
from typing import Any
from .dicom_reader import Dicom_Reader

class Analyse_Study():
    def __init__(self, config_dicom_file_validation):
        self.configuration_dicom_file_validation(config_dicom_file_validation)
        self.all_sop_instance_uids=[]
        self.study_instance_uid = None
        self.all_series_instance_uids= []
        self.series_id_dict={}

    def configuration_dicom_file_validation(self, config_dicom_file_validation: str) -> None:
        """Configuration of the attributes used in the dicom file validation.

        Args:
            config_dicom_file_validation (str): The direction to the config file.
        """
        with open(config_dicom_file_validation, 'r') as f:
            dicom_file_validation_attributes = json.load(f)
        self.file_study_instance_uid=list(dicom_file_validation_attributes.values())[0]
        self.file_sop_instance_uid=list(dicom_file_validation_attributes.values())[1]
        self.file_series_instance_uid=list(dicom_file_validation_attributes.values())[2]
        return

    def set_series(self, series_id: str, series_attributes: Any) -> None:
        """Add the attributes for each series in a dictionary containing the series id as key and the object/dictionary with attributes as value.

        Args:
            series_id (str): The string of the series id.
            series_attributes (Any): The object or dictionary containing the attributes of a series.
        """
        self.series_id_dict[series_id]=series_attributes
        return

    def get_series(self, series_id: str) -> Any:
        """Retrieves the object/dictionary containing the attributes of a series using the series id.

        Args:
            series_id (str): The string of the series id.

        Returns:
            Any: The object/dictionary containing the attributes.
        """
        return self.series_id_dict[series_id]

    def analyse_study(self, series: Dicom_Reader) -> bool:
        """Validate that all dicom files (series) of a study have the same Study Instance UID.

        Args:
            dicom_object (MetadataReader): The object that contains the dicom metadata attributes.

        Raises:
            Exception: Strings are not the same.
        
        Returns:
            bool: Returns the boolean of the assessment if the ids are duplicates.
        """
        if (self.study_instance_uid != None) and (self.study_instance_uid != series.__dict__[self.file_study_instance_uid]):
            raise Exception('Files are not from the same study')
        else:
            self.study_instance_uid = series.__dict__[self.file_study_instance_uid]

        duplicate_sop_elements= series.__dict__[self.file_sop_instance_uid] in self.all_sop_instance_uids
        duplicate_series_elements= series.__dict__[self.file_series_instance_uid] in self.all_series_instance_uids
        self.all_sop_instance_uids.append(series.__dict__[self.file_sop_instance_uid])
        self.all_series_instance_uids.append(series.__dict__[self.file_series_instance_uid])
        return duplicate_sop_elements, duplicate_series_elements


import os
import logging
import zipfile
import pydicom
from typing import Any

from .dicom_reader import Dicom_Reader
from .analyse_study import Analyse_Study
from .data_cleaning import data_cleaning_instance

class Metadata_Reader():

    def __init__(self, metadata_document_directory: str, config_dicom_file_validation) -> None:
        """Takes the path to either a single document that contains the metadata for mapping, or 
        to a folder that contains multiple corresponding files with metadata.

        Args:
            metadata_document_directory (str): String path to the metadata files.
        """
        self.analyse_study_instance=Analyse_Study(config_dicom_file_validation)
        self.all_dicom_series_dict={}
        self.all_dicom_series = []
        file_name, file_extension = os.path.splitext(
                metadata_document_directory)

        if file_extension == ".zip":
            file_name, file_extension = os.path.splitext(
                metadata_document_directory)
            with zipfile.ZipFile(metadata_document_directory) as dataset:
                for file in range(1, len(dataset.filelist)):
                    with dataset.open(dataset.filelist[file].filename) as file:
                        datasetFileName, dataset_file_extension = os.path.splitext(
                            file.name)
                        self.evaluate_file_type(
                            file, dataset_file_extension)
                
                for value in self.all_dicom_series_dict.values():
                    if len(value)>1:
                        self.all_dicom_series.extend(self.post_read_processing(value, flag="all"))
                    else:
                        self.all_dicom_series.extend(self.post_read_processing(value, flag="single"))

        elif type(file_extension) == type(str()):
            self.evaluate_file_type(metadata_document_directory, file_extension)
            self.all_dicom_series.extend(self.post_read_processing(list(self.all_dicom_series_dict.values())[0], flag="single"))
        else:
            logging.error("No valid metadata file path.")
            raise FileNotFoundError("No valid metadata file path.")
        return

    def evaluate_file_type(self, file: str, file_extension: str) -> None:
        """Takes the file path and the file extension and evaluates their type in order to call the corresponding class for metadata 
        extraction.

        Args:
            file_num (str): String path to the metadata file.
            file_extension (str): String of the file type.
        """

        if file_extension == ".dcm":
            try:
                dicom_series = Dicom_Reader(file)
                sub_dict=Dicom_Reader.pydicom_object_search(dicom_series.pydicom_file)
                dicom_series.__dict__.update(sub_dict)
                del dicom_series.pydicom_file

                duplicate_sop_elements, duplicate_series_elements=self.analyse_study_instance.analyse_study(dicom_series)
                if (duplicate_sop_elements==False) and (duplicate_series_elements==True):
                    self.all_dicom_series_dict[dicom_series.__dict__[self.analyse_study_instance.file_series_instance_uid]].append(dicom_series)
                else:
                    self.all_dicom_series_dict[dicom_series.__dict__[self.analyse_study_instance.file_series_instance_uid]]=[dicom_series]

            except pydicom.errors.InvalidDicomError as e:
                pass
            except FileNotFoundError as e:
                pass
        else:
            logging.error("File format is not supported.")
        return

    def post_read_processing(self, attributes: Any, flag):
        data_cleaning_instance.load_attributes(attributes)
        if flag=="single":
            data_cleaning_instance.set_attributes_from_config(flag="single")
        elif flag=="all":
            data_cleaning_instance.set_attributes_from_config(flag="all")
        else:
            raise("No correct flag provided.")
        return data_cleaning_instance.attributes_dict
            


class Schemas_Collector():

    def __init__(self) -> None:
        """Instantiates the class and creates an empty dictionary as attribute.
        """
        self.schemas = {}

    def add_schema(self, uri: str, schema: dict) -> None:
        """Takes as input the string of the URI that resolves to a schema and the schema as dictionary. Creates a new entry in the
        object dictionary, e.g. attribute schemas, where URI is the key and the schema dictionary is the value.

        Args:
            uri (str): The string of the schema URI.
            schema (dict): The dictionary of the JSON schema.
        """
        self.schemas[uri] = schema

    def get_uri(self, uri: str) -> bool:
        """Takes the string of the URI to a schema.

        Args:
            uri (str): The string of the schema URI.

        Returns:
            bool: Returns the boolean if the URI exists in the dictionary of the class attribute schemas.
        """
        return uri in self.schemas

    def get_schema(self, uri: str) -> dict:
        """Takes the URI for a schema as string.

        Args:
            uri (str): The string of the schema URI.

        Returns:
            dict: Returns the JSON schema from the class attribute schemas as dictionary.
        """

        return self.schemas[uri]

schemas_collector_instance = Schemas_Collector()



import json
from datetime import datetime
import os
from typing import Any
from collections.abc import Iterable

class Data_Cleaning():

    def __init__(self) -> None:
        """Instantiates the Data_Cleaning class and creates an empty attribute attribute_dict to be updatet.
        """
        self.config={}

    @classmethod
    def merge_date_time(cls, args: list, object_iterable: Any) -> dict:
        """Merge date and time values from two attributes into one attribute value.

        Args:
            args (list): Names of the original and merged attributes.
            attribute_dict (dict): The dictionry containing the key-value pairs of the attributes to be merged.

        Returns:
            dict: The updatet dictionary containing the merged attribute key-value pair.
        """
        for object in object_iterable:
            input_args=args[0]
            output_args=args[1]
            merged_date_time = object.__dict__[input_args[0]] + " "
            merged_date_time += object.__dict__[input_args[1]]
            try:
                merged_date_time = datetime.strptime(
                    merged_date_time, '%Y%m%d %H%M%S').isoformat()
            except:
                merged_date_time=merged_date_time
            object.__dict__.pop(input_args[0])
            object.__dict__.pop(input_args[1])
            object.__dict__[output_args]=merged_date_time
        return object_iterable

    @classmethod
    def merge_min_max_values(cls, args: list, object_iterable: Any) -> dict:
        input_args=args[0]
        min_value=min([object.__dict__[input_args[0]] for object in object_iterable])
        max_value=max([object.__dict__[input_args[1]] for object in object_iterable])
        for object in object_iterable:
            object.__dict__[input_args[0]]=min_value
            object.__dict__[input_args[1]]=max_value
        return object_iterable
    
    @classmethod
    def transfer_to_list(cls, iterable_atribute: Iterable) -> list:
        """Transfers an iterable attribute to a list.

        Args:
            iterable_atribute (Iterable): The iterable attribute to be converted.

        Returns:
            list: The resulting list.
        """
        new_list=[]
        for value in iterable_atribute:
            if isinstance(value, list):
                new_list.append(cls.iterate_list(value))
            else:
                new_list.append(value)
        return new_list

    def set_attributes_from_config(self, flag, config_file_path: str=os.getcwd()+"/configs/config_data_cleaning.json",) -> None:
        """Uses the JSON config file for the data cleaning to assess the methods and respective attributes that need to be cleaned. Attributes 
        and methods are then iteratively called and executed. The object instance attribute attribute_dict is then updatet after each process.

        Args:
            config_file_path (str, optional): The relative path to the config file directory. Defaults to os.getcwd()+"/configs/config_data_cleaning.json".
        """
        with open(config_file_path, 'r') as f:
            config_file = json.load(f)
        for func_string, args in config_file.items():
            self.store_config(func_string, args)
            func = getattr(Data_Cleaning, func_string)
            temp_attribute_dict=func(args, self.attributes_dict)
            self.attributes_dict=temp_attribute_dict
        return

    def load_attributes(self, attributes_dict: dict) -> None:
        """Load a dictionary to update the attribute_dict object instance attribute.

        Args:
            attributes_dict (dict): The dictionary containing the key-value pairs for the data cleaning setps.
        """
        self.attributes_dict=attributes_dict
        return
    
    def store_config(self, func_string, args):
        self.config.update({func_string: args})
        return
    
data_cleaning_instance=Data_Cleaning()



import logging
from typing import Any

class Schema_Reader():

    def __init__(self, schema: dict) -> None:
        """Instantiates the class that searches the schema structure of the provided dictionary and builds a schema skeleton. When instantiated, the json definitions are first set.

        Args:
            schema (dict): The json schema document as dictionary.
        """
        if "$defs" in schema.keys():
            self.definitions = schema["$defs"]
        elif "definitions" in schema.keys():
            self.definitions = schema["definitions"]
        else:
            self.definitions = None
        self.schema = schema

    def json_definition_search(self, definition: dict) -> dict:
        """Takes a dictionary as input that represents the reference to a json definition as content. Returns the skeleton of this content as a dictionary.

        Args:
            definition (dict): The reference content as dictionary.

        Returns:
            dict: The skeleton of the reference content as dictionary.
        """
        properties = None
        if "$ref" in definition:
            if definition["$ref"].startswith("#"):
                keyword = definition["$ref"].split("/")[-1:][0]
                sub_properties = self.json_definition_search(self.definitions[keyword])
                properties = sub_properties
            else:
                path = definition["$ref"]
                logging.warning("No correct definition path for " + path)
        elif "oneOf" in definition:
            sub_properties=self.one_of_search(definition)
            properties = sub_properties
        elif definition["type"] == "array":
            sub_properties = self.json_array_search(definition["items"])
            properties = sub_properties
        elif definition["type"] == "object":
            sub_properties = self.json_object_search(definition)
            properties = sub_properties
        else:
            properties = self.json_type_search(definition["type"])
        return properties

    def json_array_search(self, property: dict) -> list:
        """Takes a dictionary as input that represents the array type content of a json document. Returns the skeleton of this content as a list.

        Args:
            property (dict): The array type content as dictionary.

        Returns:
            list: The skeleton of the array type content as list.
        """
        properties = None
        if "$ref" in property:
            if property["$ref"].startswith("#"):
                keyword = property["$ref"].split("/")[-1:][0]
                sub_properties = [self.json_definition_search(self.definitions[keyword])]
                properties = sub_properties
            else:
                path = property["$ref"]
                logging.warning("No correct definition path for " + path)
        elif "oneOf" in property:
            sub_properties=self.one_of_search(property)
            properties = sub_properties
        elif property["type"] == "array":
            sub_properties = [self.json_array_search(property["items"])]
            properties = sub_properties
        elif property["type"] == "object":
            sub_properties = [self.json_object_search(property)]
            properties = sub_properties
        else:
            sub_properties = [self.json_type_search(property["type"])]
            properties = sub_properties
        return properties

    def json_object_search(self, property: dict) -> dict:
        """Takes a dictionary as input that represents the object type content of a json document. Returns the skeleton of this content as a dictionary.

        Args:
            property (dict): The object type content as dictionary.

        Returns:
            dict: The skeleton of the object type content as list.
        """
        properties = {}
        for i in property["properties"].items():
            if "$ref" in i[1]:
                if i[1]["$ref"].startswith("#"):
                    keyword = i[1]["$ref"].split("/")[-1:][0]
                    sub_properties = self.json_definition_search(self.definitions[keyword])
                    properties[i[0]] = sub_properties
                else:
                    path = i[1]["$ref"]
                    logging.warning("No correct definition path for " + path)
            elif "oneOf" in i[1]:
                sub_properties=self.one_of_search(i[1])
                properties[i[0]] = sub_properties
            elif i[1]["type"] == "array":
                sub_properties = self.json_array_search(i[1]["items"])
                properties[i[0]] = sub_properties
            elif i[1]["type"] == "object":
                sub_properties = self.json_object_search(i[1])
                properties[i[0]] = sub_properties
            else:
                if i[0] == "value":
                    properties[i[0]] = self.json_type_search(i[1]["type"])
                elif i[0] == "unit":
                    try:
                        properties[i[0]] = i[1]["default"]
                    except KeyError:
                        logging.warning("No default unit")
                        properties[i[0]] = self.json_type_search(i[1]["type"])
                else:
                    properties[i[0]] = self.json_type_search(i[1]["type"])
        return properties

    def json_type_search(self, type: str) -> Any:
        """Takes a string as input that represents the type of a json document, i.e. a primitive data type or a string. 
        The input string is altered and returned either as a string, or a list, if multiple types are present 

        Args:
            type (str): The string of the json type.

        Returns:
            Any: The string of the Python type, or a list of Python types as strings.
        """
        if type == "integer":
            return "<class 'int'>"
        elif type == "string":
            return "<class 'str'>"
        elif type == "number":
            return "<class 'float'>"
        elif type == "boolean":
            return "<class 'bool'>"
        elif type == "null":
            return "<class 'NoneType'>"
        elif type == "array":
            return "<class 'list'>"
        elif isinstance(type, list):
            multiple_types = []
            for j in type:
                multiple_types.append(self.json_type_search(j))
            return tuple(multiple_types)
        else:
            logging.warning("Type Error")
            return None
    
    def one_of_search(self, property: dict) -> list:
        """Takes a dictionary as input that represents the oneOf type content of a json document, which is an array of multiple possible values. 
        Returns the skeleton of this content as a list.

        Args:
            property (dict): The oneOf type content as dictionary.

        Returns:
            list: The skeleton of the oneOf type content as list.
        """
        sub_properties=[]
        for i in property["oneOf"]:
            if "items" in i:
                sub_sub_properties = self.json_array_search(i["items"])
                sub_properties.append(sub_sub_properties)
                properties=sub_properties
            elif "properties" in i:
                sub_sub_properties = self.json_object_search(i)
                sub_properties.append(sub_sub_properties)
                properties=sub_properties
            elif "type" in i: 
                sub_sub_properties = self.json_type_search(i["type"])
                sub_properties.append(sub_sub_properties)
                properties=sub_properties
            else:
                sub_properties=[None]
                properties=sub_properties
                logging.warning("No correct definition for \"oneOf\" attribute")
        return properties




from .dicom_mapping import Dicom_Mapping
from .dicom_reader import Dicom_Reader
from .attribute_inserter import Attribute_Inserter
from .metadata_reader import Metadata_Reader
from .schema_reader import Schema_Reader
from .attribute_mapper import Attribute_Mapper
from .cache_schemas import Cache_Schemas
from .schemas_collector import Schemas_Collector

__all__ = [
    'Dicom_Mapping',
    'Dicom_Reader',
    'Attribute_Inserter',
    'Map_MRI_Schema',
    'Schema_Reader',
    'Attribute_Mapper',
    'Cache_Schemas',
    'Cache_Schemas',
    'Metadata_Reader',
    'Schemas_Collector'
]


import json
import os
import logging
from .schema_reader import Schema_Reader
from .dicom_reader import Dicom_Reader
from .cache_schemas import Cache_Schemas
from .analyse_study import Analyse_Study
from .data_cleaning import data_cleaning_instance
from .metadata_reader import Metadata_Reader
from .attribute_mapper import Attribute_Mapper
from .attribute_inserter import Attribute_Inserter

class Dicom_Mapping():
    
    def __init__(self, map_json_path: str, metadata_files_location: str, mapped_metadata: str='mapped_metadata.json', config_dicom_file_validation: str=os.getcwd()+"/configs/config_dicom_file_validation.json") -> None:
        """Instantiates the class, loads the map dictionary from JSON, instantiates all attributes to the object and executes the steps for mapping.

        Args:
            map_json (json): A json based map of the attribute assignments for mapping.
            metadata_files_location (str): The directory where the dicom files of a study are stored.
            mapped_metadata (str, optional): The resulting json file. Defaults to 'mapped_metadata.json'.
        """
        with open(map_json_path, 'r') as f:
            map_dict = json.load(f)
        self.map_dict=map_dict
        self.metadata_files_location = metadata_files_location
        self.mapped_metadata=mapped_metadata
        self.config_dicom_file_validation=config_dicom_file_validation
        self.analyse_study_instance=Analyse_Study(config_dicom_file_validation)
        data_cleaning_instance
        self.execute_steps(map_dict, metadata_files_location, mapped_metadata)

    def execute_steps(self, map_dict: dict, metadata_files_location: str, mapped_metadata: str) -> None:
        """Executes all steps for mapping a dicom study to a json schema, i.e. download or cache the target schema, extract the schema skeleton, read the metadata from the DICOM files and extract metadata,
        map the attributes provided in the JSON metadata map, create the mapped metadata object, fill the schema skeleton with the attributes in the mapped metadata object and finally store the result as JSON document.

        Args:
            map_dict (dict): The map of the attribute assignments for mapping as a dictionary.
            metadata_files_location (str): The directory where the dicom files of a study are stored.
            mapped_metadata (str): The resulting json file.
        """
        json_schema = Cache_Schemas.cache_schema(map_dict).json_schema
        schema_skeleton = Schema_Reader(json_schema)
        self.schema_skeleton = schema_skeleton.json_object_search(schema_skeleton.schema)
        dicom_object = Metadata_Reader(metadata_files_location, self.config_dicom_file_validation)
        dicom_series_list = dicom_object.all_dicom_series
        study_map = Attribute_Mapper.mapping_from_object(dicom_series_list[0].__dict__, map_dict, list(map_dict.keys())[1], [])
        if len(list(map_dict.keys())) > 2:
            merged_series_dict=self.map_and_merge_series(dicom_series_list, map_dict)
            for merged_series in merged_series_dict.values():

                if list(map_dict.keys())[1] in study_map:
                    merged_study_map=Attribute_Mapper.merge_mapped_attributes(study_map[list(map_dict.keys())[1]], merged_series, list(map_dict.keys())[2])
                else:
                    merged_study_map=Attribute_Mapper.merge_mapped_attributes(study_map, merged_series, list(map_dict.keys())[2])
            merged_study_map={list(map_dict.keys())[1]: merged_study_map}
        else:
            pass
        
        map_schema = Attribute_Inserter(self.schema_skeleton, list(self.schema_skeleton.keys()), merged_study_map)
        filled_schema = map_schema.fill_json_object(map_schema.schema_skeleton, map_schema.key_list, map_schema.map)
        with open(mapped_metadata, 'w') as f:
            json.dump(filled_schema, f)

        return

    def map_and_merge_series(self, dicom_series_list: list, map_dict: dict) -> dict:
        """Mapping and merging of the series that are part of a MRI study.

        Args:
            dicom_series_list (list): The list of metadata objects from the metadata extraction of the DICOM files.
            map_dict (dict): The JSON map that contains the attribute assignments of the origin and target attributes.

        Returns:
            dict: The dictionary that contains the mapped and merged metadata attributes.
        """
        merged_series_dict={}
        if len(list(map_dict.keys())) == 4:
            for raw_series in dicom_series_list:
                duplicate_sop_elements, duplicate_series_elements=self.analyse_study_instance.analyse_study(raw_series)
                if (duplicate_sop_elements==False) and (duplicate_series_elements==True):
                    series_map={list(map_dict.keys())[2]: self.analyse_study_instance.get_series(raw_series.__dict__[self.analyse_study_instance.file_series_instance_uid])}####
                    merged_series_map=self.series_extension(map_dict, list(map_dict.keys())[3], raw_series, series_map)
                elif (duplicate_sop_elements==True) and (duplicate_series_elements==True):
                    merged_series_map=Attribute_Mapper()
                    pass
                else:
                    series_map = Attribute_Mapper.mapping_from_object(raw_series.__dict__, map_dict, list(map_dict.keys())[2], [list(map_dict.keys())[1]])
                    merged_series_map=self.series_extension(map_dict, list(map_dict.keys())[3], raw_series, series_map)
                    self.analyse_study_instance.set_series(raw_series.__dict__[self.analyse_study_instance.file_series_instance_uid], merged_series_map)
                merged_series_dict[raw_series.__dict__[self.analyse_study_instance.file_series_instance_uid]]=merged_series_map 
        else:
            for raw_series in dicom_series_list:
                duplicate_sop_elements, duplicate_series_elements=self.analyse_study_instance.analyse_study(raw_series)
                if (duplicate_sop_elements==True) and (duplicate_series_elements==True):
                    series_map=Attribute_Mapper()
                    pass
                else:
                    series_map = Attribute_Mapper.mapping_from_object(raw_series.__dict__, map_dict, list(map_dict.keys())[2], [list(map_dict.keys())[1]])
                    self.analyse_study_instance.set_series(raw_series.__dict__[self.analyse_study_instance.file_series_instance_uid], series_map)
                merged_series_dict[raw_series.__dict__[self.analyse_study_instance.file_series_instance_uid]]=series_map

        return merged_series_dict

    def series_extension(self, map_dict: dict, map_attribute: str, series: Dicom_Reader, series_map) -> list:
        """Extends the mapped attributes of a series object by an attribute that has a list of objects as values, using the keywords of the provided map.

        Args:
            map_dict (dict): Map that contains the attribute assignments for the dicom metadata and the schema.
            map_attribute (str): The attribute in the map that contains the mapping assignments.
            series (DicomReader): The series which is extended.

        Returns:
            list: A list of objects with the mapped attributes.
        """
        attribute_value=None
        target_attributes={}
        for origin_attribute in list(map_dict[map_attribute].values()):
            try:
                attribute_value=series.__dict__[origin_attribute]
            except KeyError as e:
                continue
            target_attributes[origin_attribute]=attribute_value
        attributes_map = Attribute_Mapper.mapping_from_object(target_attributes, map_dict, map_attribute, [list(map_dict.keys())[1], list(map_dict.keys())[2]])

        assess_type=Attribute_Mapper(**{"schema_skeleton": self.schema_skeleton})
        all_types, number_of_elements=assess_type.type_assessment(attributes_map, map_dict, map_attribute)

        for key in all_types:
            new_attributes_map={}
            if not key in attributes_map:
                new_attributes_map.update(assess_type.nested_attributes_map_search(attributes_map, key))
            else:
                pass

        if len(new_attributes_map) > 0:
            pass
        else:
            new_attributes_map=attributes_map

        for element in range(0, number_of_elements):
            element_series_attribute={}

            for series_attribute, attribute_value in new_attributes_map.items():
                if (isinstance(attribute_value, list)) and (((isinstance(attribute_value[0], list)) and (str(type(attribute_value[0]))== all_types[series_attribute])) or ((isinstance(attribute_value[0], list))==False) and (str(type(attribute_value[0])) == all_types[series_attribute])):    
                    element_series_attribute[series_attribute]=attribute_value[element]

                elif isinstance(attribute_value, list) == False:
                    element_series_attribute[series_attribute]=element_series_attribute[series_attribute]=attribute_value

                elif (isinstance(attribute_value, list)) and (str(type(attribute_value)) == all_types[series_attribute]):
                    element_series_attribute[series_attribute]=element_series_attribute[series_attribute]=attribute_value
                    
                else:
                    logging.error(f"The value for attribute {series_attribute} does not correspond to the required value in the schema.")

            if set(element_series_attribute.keys()).issubset(set(attributes_map.keys())):
                pass
            else:
                element_series_attribute=assess_type.nested_attributes_map_modification(attributes_map, element_series_attribute)

            if list(map_dict.keys())[2] in series_map:
                merged_series_map=assess_type.merge_mapped_attributes(series_map[list(map_dict.keys())[2]], element_series_attribute, list(map_dict.keys())[3])
            else:
                merged_series_map=assess_type.merge_mapped_attributes(series_map, element_series_attribute, list(map_dict.keys())[3])

        return merged_series_map    


import pydicom
import re
import logging
from .data_cleaning import data_cleaning_instance

class Dicom_Reader():

    def __init__(self, dicom_file: str) -> None:
        """Takes a dicom file as input and converts the file into a Python object using the pydicom module.

        Args:
            dicom_file (str): String that contains the location to the dicom file.
        """
        try:
            self.pydicom_file = pydicom.dcmread(dicom_file)
        except pydicom.errors.InvalidDicomError as e:
            logging.error("InvalidDicomError for file: %s %s", dicom_file, e)
            raise
        except FileNotFoundError as e:
            logging.error("FileNotFoundError for file: %s %s", dicom_file, e)
            raise
        
    @classmethod
    def pydicom_object_search(cls, dataset: pydicom) -> dict:
        """Takes as input a pydicom object and searches its attributes. Puts the attributes which contain a string as value, or a list of values into a dictionary.

        Args:
            dataset (pydicom): The pydicom object that contains all attributes of a dicom file in a nested structure.

        Returns:
            dict: The dictionary that contains the dicom attributes in a flat structure as key-value pairs.
        """
        sub_dict = {}
        for attribute in dataset:
            if isinstance(attribute, pydicom.Dataset):
                sub_dict=cls.pydicom_object_search(attribute)
            elif isinstance(attribute.value, pydicom.Sequence):
                name=cls.name_standardization(attribute.name)
                if len(attribute.value) > 1:
                    mergedsub_dict={}
                    for value in attribute.value:
                        subsub_dict=(cls.pydicom_object_search(value))
                        mergedsub_dict=cls.merge_dict_keys(subsub_dict, mergedsub_dict)
                    sub_dict.update(mergedsub_dict)
                    
                else:
                    sub_dict.update(cls.pydicom_object_search(attribute.value))
            else:
                name=cls.name_standardization(attribute.name)
                if isinstance(attribute.value, pydicom.multival.MultiValue):
                    sub_dict[name]=data_cleaning_instance.transfer_to_list(attribute.value)
                elif isinstance(attribute.value, pydicom.valuerep.PersonName):
                        sub_dict[name]=str(attribute.value)
                elif isinstance(attribute.value, pydicom.uid.UID):
                        sub_dict[name]=str(attribute.value)
                elif isinstance(attribute.value, pydicom.valuerep.DSfloat):
                        sub_dict[name]=str(attribute.value)
                elif isinstance(attribute.value, pydicom.valuerep.IS):
                        sub_dict[name]=str(attribute.value)
                else:
                    sub_dict[name] = attribute.value
        return sub_dict

    @classmethod
    def name_standardization(cls, attribute: str) -> str: 
        """Takes a string of a dicom attribute as input and standardizes it after defined criteria.

        Args:
            attribute (str): The attribute string that should be standardized.

        Returns:
            str: The attribute string after standardization.
        """
        name = attribute.split()
        if len(name) == 1:
            name = name[0].lower()
        else:
            subname = ""
            for letter in name[1:]:
                subname += letter.capitalize()
            name = name[0].lower() + subname
        name = re.sub('[^A-Za-z0-9]+', '', name)
        return name

    @classmethod
    def merge_dict_keys(cls, subsub_dict: dict, mergedsub_dict: dict) -> dict:
        """Takes as input a two dictionaries that contain attributes of a repeating attribute sequence in the pydicom file. It merges the file contains and returns one dictionary that
        has a list of values from both dictionaries for each key in the new dictionary.

        Args:
            subsub_dict (dict): The dictionary with the new values for each key.
            mergedsub_dict (str): The dictioanry with the current values in a list for each key.

        Returns:
            dict: The merged dictionary of both input dictionaries.
        """
        new_mergedsub_dict={}
        dict_keys = [key for key in subsub_dict.keys()]
        try:
            for key in dict_keys:
                if isinstance(mergedsub_dict[key], list):
                    if (isinstance(subsub_dict[key], list)) and (isinstance(mergedsub_dict[key][0], list)==False):
                        mergedsub_dict[key]=[mergedsub_dict[key]]
                    else:
                        pass
                    mergedsub_dict[key].append(subsub_dict[key])
                    new_mergedsub_dict[key] = mergedsub_dict[key]
                else:
                    new_mergedsub_dict[key] = [mergedsub_dict[key], subsub_dict[key]]
        except KeyError as e:
            new_mergedsub_dict = subsub_dict
        return new_mergedsub_dict


from typing import Any
import logging
from nested_lookup import nested_lookup

class Attribute_Mapper():

    def __init__(self, **kwargs) -> None:
        """Instantiates the class and updates the attributes.
        """
        if kwargs:
            self.__dict__.update(kwargs)
        else:
            pass
    
    def update_object_attributes(self, attributes: dict) -> None:
        """Updates the key-value pairs of the object attributes dictionary.

        Args:
            attributes (dict): Attributes to be added to the object dictionary.
        """
        self.__dict__.update(attributes)
        
        return 
    
    @classmethod
    def mapping_from_object(cls, metadata_attributes: dict, map_dict: dict, map_main_attribute: str, object_names: list) -> dict:
        """Takes the metadata object as dictionary and the corresponding map dictionary to insert the proper values for a main attribute from the
        map dictionary.

        Args:
            metadata_attributes (dict): Dictionary of attributes from the metadata object.   
            map_dict (dict): Dictionary of the map for the origin and target schemas.
            map_main_attribute (str): The main attribute of the map dictionary (usually study, series and perImage) for which all attributes should be mapped.

        Returns:
            dict: The dictionary with the mapped attributes. Passes to the __init__ function, which updates the attributes.
        """
        mapped_attribues={}
        for key, value in map_dict[map_main_attribute].items():
            if value in metadata_attributes:
                if "." in key:
                    split_keys=key.split(".")
                    split_keys.append(metadata_attributes[value])
                    split_keys=cls.remove_from_list(split_keys, object_names)
                    for index in range(len(split_keys)-1, 1, -1):
                        temp_key_dict={}
                        temp_key_dict[split_keys[index-1]]= split_keys[index]
                        split_keys.remove(split_keys[index])
                        split_keys.remove(split_keys[index-1])
                        split_keys.append(temp_key_dict)
                    if split_keys[0] in mapped_attribues:
                        merged_attribute_object=cls.merge_mapped_attributes(mapped_attribues[split_keys[0]], temp_key_dict)
                        mapped_attribues[split_keys[0]]=merged_attribute_object
                    else:
                        attribute_object=Attribute_Mapper()
                        attribute_object.update_object_attributes(temp_key_dict)
                        mapped_attribues[split_keys[0]]=attribute_object
                else:
                    mapped_attribues[key]= metadata_attributes[value]
            else:
                pass
        return mapped_attribues

    @classmethod
    def merge_mapped_attributes(cls, existing_attributes: Any, new_attributes: Any, list_attribute=None) -> object:
        """Merges the mapped attributes that belong to the same hirarchy in the target schema, by adding them to the same object.

        Args:
            existing_attributes (Any): The object, or dictionary containing the mapped attributes of the same schema hierarchy.
            new_attributes (Any): The object, or dictionary containing new attributes that need to be merged with the (object) dictionary of mapped attributes.
            list_attribute (_type_, optional): In case a list attribute exists that needs to be added to the attribute dictionary. Defaults to None.

        Returns:
            object: The object containing the merged new and existing mapped attributes.
        """
        temp_dict={}
        if isinstance(existing_attributes, Attribute_Mapper):
            attribute_dict=existing_attributes.__dict__
        else:
            attribute_dict=existing_attributes
        if isinstance(new_attributes, Attribute_Mapper):
            new_attributes=new_attributes.__dict__
        else:
            pass
        if list(new_attributes.keys())[0] == list_attribute:
            if list_attribute in attribute_dict:
                attribute_dict[list_attribute].append(new_attributes[list_attribute])
            else:
                attribute_dict[list_attribute]=[new_attributes[list_attribute]]
            attribute_object_1=Attribute_Mapper()
            attribute_object_1.update_object_attributes(attribute_dict)
            return attribute_object_1
        elif list(new_attributes.keys())[0] in (attribute_dict.keys()):
            attribute_dict[list(new_attributes.keys())[0]]=cls.merge_mapped_attributes(attribute_dict[list(new_attributes.keys())[0]], new_attributes[list(new_attributes.keys())[0]], list_attribute)
            if not isinstance(attribute_dict, Attribute_Mapper):
                attribute_object_1=Attribute_Mapper()
                attribute_object_1.update_object_attributes(attribute_dict)
            else:
                temp_dict[list(new_attributes.keys())[0]]=attribute_dict
                attribute_object_1=Attribute_Mapper()
                attribute_object_1.update_object_attributes(temp_dict)
            return attribute_object_1
        elif list_attribute!=None:
            if list_attribute in attribute_dict:
                attribute_object_1=Attribute_Mapper()
                attribute_object_1.update_object_attributes(new_attributes)
                attribute_dict[list_attribute].append(attribute_object_1)
            else:
                attribute_object_1=Attribute_Mapper()
                attribute_object_1.update_object_attributes(new_attributes)
                attribute_dict[list_attribute]=[attribute_object_1]
            attribute_object_2=Attribute_Mapper()
            attribute_object_2.update_object_attributes(attribute_dict)
            return attribute_object_2
        else:
            attribute_dict.update(new_attributes)
            attribute_object_1=Attribute_Mapper()
            attribute_object_1.update_object_attributes(attribute_dict)
            return attribute_object_1

    @classmethod
    def remove_from_list(cls, list1: list, list2: list) -> list:
        """Compares the values between list1 and list2 and appends matching values in a new list which is returned.

        Args:
            list1 (list): The first list.
            list2 (list): The second list.

        Returns:
            list: The list that contains the values which are contained in both lists.
        """
        new_list=[]
        for element in list1:
            if element in list2:
                pass
            else:
                new_list.append(element)
        return new_list

    def get_data_type(self, target_attribute: str, split_target_attribute: str=None, main_target_attribute: str=None) -> Any:
        """Looks up the schema in the object instance to retreive the value of the target attribute provided to the function.

        Args:
            target_attribute (str): The attribute of which the value should be looked up in the schema.
            split_target_attribute (str, optional): The attribute from the end of a string representing a sequence of attributes refering to the schema hierarchy. Defaults to None.
            main_target_attribute (str, optional): In case of a string representing a sequence of attributes, this value represents the original attribute of which value has to be searched in the schema. Defaults to None.

        Returns:
            Any: The value of the attribute in the schema and the attribute.
        """
        
        if "." in target_attribute:
            split_target_attribute=target_attribute.split(".")
            main_target_attribute=split_target_attribute[-1]
            attribute_type=nested_lookup(split_target_attribute[-1], self.schema_skeleton)
        else:
            attribute_type=nested_lookup(target_attribute, self.schema_skeleton)
            main_target_attribute=target_attribute

        if len(attribute_type)>1:
            try:
                split_target_attribute.pop()
                main_target_attribute=target_attribute
                attribute_type, main_target_attribute=self.get_data_type(split_target_attribute[-1], split_target_attribute, main_target_attribute)#check again main attribute target
            except TypeError as e:
                logging.error(f"attribute {attribute_type[0]} occurs multiple times and full path is not provided in the metadata map.")
            except AttributeError as e:
                logging.error(f"attribute {attribute_type[0]} occurs multiple times and full path is not provided in the metadata map.")
        else:
            pass
        if isinstance(attribute_type[0], dict):
            attribute_type=nested_lookup(main_target_attribute, attribute_type[0])
            return attribute_type[0]
        else:
            return attribute_type[0], main_target_attribute

    @classmethod
    def nested_list_level(cls, lst: list) -> int:
        """Returns the nesting level of a list.

        Args:
            lst (list): The nested list.

        Returns:
            int: The level of the list nesting.
        """
        if isinstance(lst, list):
            return 1 + max(cls.nested_list_level(item) for item in lst)
        else:
            return 0

    @classmethod
    def nested_attributes_map_search(cls, attributes_map: dict, target_attribute: str) -> Any:
        """Searches a nested dictionary for a certain target attribute.

        Args:
            attributes_map (dict): The nested dictionary to be searched.
            target_attribute (str): The target attribute in the dictionary.

        Returns:
            Any: The level of the dictionary where the attribute is located at.
        """
        for attribute in attributes_map:
            if (attribute==target_attribute) or target_attribute in attributes_map:
                return attributes_map
            elif isinstance(attributes_map[attribute], Attribute_Mapper):
                return cls.nested_attributes_map_search(attributes_map[attribute].__dict__, target_attribute)
            elif isinstance(attributes_map[attribute], dict):
                return cls.nested_attributes_map_search(attributes_map[attribute], target_attribute)
            else:
                logging.warning(f"{target_attribute} not found in {attributes_map}.")
                pass
        return attributes_map

    @classmethod
    def nested_attributes_map_modification(cls, attributes_map, element_series_attribute):
        for attribute in attributes_map:
            if attribute in element_series_attribute:
                attributes_map[attribute]=element_series_attribute[attribute]
            elif isinstance(attributes_map[attribute], Attribute_Mapper):
                attributes_map[attribute].__dict__.update(cls.nested_attributes_map_modification(attributes_map[attribute].__dict__, element_series_attribute))
                return attributes_map
            elif isinstance(attributes_map[attribute], dict):
                attributes_map[attribute].update(cls.nested_attributes_map_modification(attributes_map[attribute], element_series_attribute))
                return attributes_map
            else:
                logging.warning(f"{element_series_attribute} not found in {attributes_map}.")
                pass
        return attributes_map

    def type_assessment(self, original_attributes_map, map_dict, map_attribute):
        number_of_elements=0
        all_types={}
        for target_attribute in map_dict[map_attribute]:
            attribute_type, target_attribute=self.get_data_type(target_attribute)
            if not target_attribute in original_attributes_map:
                attributes_map=self.nested_attributes_map_search(original_attributes_map, target_attribute)
                if not target_attribute in attributes_map:
                    continue
                else:
                    pass
            else:
                attributes_map=original_attributes_map
            if (isinstance(attributes_map[target_attribute], list)) and (self.nested_list_level(attributes_map[target_attribute]) > self.nested_list_level(attribute_type)):
                number_of_elements=len(attributes_map[target_attribute])
            elif (isinstance(attributes_map[target_attribute], list)) and (self.nested_list_level(attributes_map[target_attribute]) == self.nested_list_level(attribute_type)):
                number_of_elements=1
            else:
                logging.warning(f"no correct type for attribute: {target_attribute}")
            if isinstance(attribute_type, list):
                attribute_type=str(list)
            elif isinstance(attribute_type, dict):
                attribute_type=str(dict)
            elif isinstance(attribute_type, tuple):
                attribute_type=str(dict)
            else:
                pass
            all_types[target_attribute]=attribute_type
        return all_types, number_of_elements


from typing import Any
import logging
from .attribute_mapper import Attribute_Mapper


class Attribute_Inserter():

    def __init__(self, schema_skeleton: dict, key_list: list, map: object) -> None:
        """Takes the schema structure as dictionary and the metadata map to assign the metadata values to the schema attributes
        at the proper hirarchy level of the JSON schema.

        Args:
            schema (dict): The dictionary containing the skeleton of the target schema.
            key_list (list): The list that contains all keys of the schema skeleton at the first hierarchy level.
            map (object): The object that represents the mapped attributes.
        """
        self.schema_skeleton = schema_skeleton
        self.key_list = key_list
        self.map = map

    def fill_json_object(self, json_object: dict, key_list: list, attributes_object: object):
        """Takes the JSON object of the schema structure, the key list of the current schema hierarchy level and the attributes object. 
        Calls the proper method to insert the attribute value, based on the attribute type. Has in addition a functionality for inserting
        the value of the main key string of an JSON object containing value and unit properties.

        Args:
            json_object (dict): The dictionary containing the skeleton of the target schema.
            key_list (list): The list that contains all keys of the schema skeleton at the first hierarchy level.
            attributes_object (object): The object that contains the mapped attributes.

        Returns:
            dict: The dictionary that represents the filled JSON object.
        """
        new_dict = {}

        for key in key_list:
            if (isinstance(json_object[key], str)) or (isinstance(json_object[key], tuple)):
                if key in attributes_object.keys():
                    new_dict[key] = self.get_json_type(json_object[key], attributes_object[key], key)
                else:
                    new_dict[key] = self.get_json_type(json_object[key], json_object[key], key)
            elif isinstance(json_object[key], dict):
                if key in attributes_object:
                    if isinstance(attributes_object[key], Attribute_Mapper):
                        sub_dict = self.fill_json_object(json_object[key], list(json_object[key].keys()), attributes_object[key].__dict__)
                    else:
                        sub_dict = self.fill_json_object(json_object[key], list(json_object[key].keys()), attributes_object[key])
                else:
                    sub_dict = self.fill_json_object(json_object[key], list(json_object[key].keys()), attributes_object)
                if len(sub_dict) > 0:
                    new_dict[key] = sub_dict
            elif isinstance(json_object[key], list):
                filled_array=[]
                if key in attributes_object:
                        filled_array = self.fill_json_array(json_object, key, json_object[key], attributes_object[key])
                else:
                    pass
                if len(filled_array) > 0:
                    new_dict[key] = filled_array
                else:
                    pass
            else:
                pass
        return new_dict

    def fill_json_array(self, json_object: dict, json_object_property: str, json_array: list, attributes: Any) -> list:
        """Takes the json object of the schema structure, the property that contains the json array as value, the json array and the attributes, either as object, or list of objects.
        Parses the array in the schema structure and calls the proper method to insert the attribute value, based on the attribute type.

        Args:
            json_object (dict): The dictionary that represents the schema skeleton object.
            json_object_property (str): The string property that contains the json array as value.
            json_array (list): The list that represents the schema skeleton array.
            attributes (Any): The object that contains the mapped attributes, or a list of objects that contain the mapped attributes.

        Returns:
            list: The list that represents the filled json array.
        """
        if not isinstance(attributes, list):
            if json_object_property in attributes.__dict__.keys():
                attributes = attributes.__dict__[json_object_property]
        try:
            json_array = json_array * len(attributes)
        except TypeError as e:
            logging.warning(e)
            pass

        new_list = []

        for list_item, list_index in zip(json_array, range(0, len(json_array))):
            if (isinstance(list_item, str)) or (isinstance(list_item, tuple)):
                try:
                    new_list.append(self.get_json_type(list_item, attributes[list_index], None))
                except TypeError as e:
                    logging.warning(e)
                    pass
            elif isinstance(list_item, dict):
                if isinstance(attributes[list_index], Attribute_Mapper):
                    new_list.append(self.fill_json_object(list_item, list(
                    list_item.keys()), attributes[list_index].__dict__))
                else:
                    new_list.append(self.fill_json_object(list_item, list(
                        list_item.keys()), attributes[list_index]))
            elif isinstance(list_item, list):
                new_list.append(self.fill_json_array(json_object, json_object_property, list_item, attributes[list_index]))
            else:
                pass
        return new_list

    def get_json_type(self, data_type: str, attribute: str, key) -> Any:
        """Takes an attribute and its data type. Confirms the primitive data types of the mapped attribute values and assigns these values to the schema attribute. The correct hirarchial
        position has been reached through the methods above.

        Args:
            data_type (str): The string that representes the data type for an attribute.
            attribute (str): The string that represents the value of the mapped attribute.

        Returns:
            Any: The value of the mapped attribute as the correct data type.
        """
        try:
            if (isinstance(data_type, tuple) and not (isinstance(attribute, tuple))):
                if (isinstance(attribute, list)) and ("<class 'list'>" in data_type):
                    return attribute
                else:
                    for element in data_type:
                        if element in ["<class 'int'>", "<class 'bool'>", "<class 'None'>", "<class 'float'>", "<class 'list'>"]:
                            return (self.get_json_type(element, attribute, None))
                        else:
                            pass
                    logging.warning(f'incorrect type provided for property \"{key}\", expected {data_type} but received {type(attribute)}, returning original value.')
                    return attribute
            elif isinstance(attribute, tuple):
                logging.warning(f'No value provided for property \"{key}\", returning original value.')
                return attribute
            elif data_type == "<class 'int'>":
                return int(attribute)
            elif data_type == "<class 'bool'>":
                return bool(attribute)
            elif data_type == "<class 'None'>":
                return "null"
            elif data_type == "<class 'float'>":
                return float(attribute)
            elif data_type == "<class 'str'>":
                return str(attribute)
            else:
                logging.warning(f'incorrect type provided for property \"{key}\", expected {data_type} but received {type(attribute)}, returning original value.')
                return attribute
        except TypeError:
            logging.warning(f'incorrect type provided for property \"{key}\", expected {data_type} but received {type(attribute)}, returning original value.')
            return attribute
        except ValueError:
            logging.warning(f'incorrect value provided for property \"{key}\", expected {data_type} but received {type(attribute)}, returning original value.')
            return attribute


from distutils.core import setup
setup(
  name = 'jammato',        
  packages = ['jammato'],   
  version = '1.3',      
  license='cc',        
  description = 'A package for mapping metadata attributes from proprietary file format schemas to a json schema document.',
  author = 'Nicolas Blumenroehr',                  
  author_email = 'nicolas.blumenroehr@gmail.com',      
  url = 'https://github.com/user/nicolasblumenroehr',   
  download_url = 'https://github.com/kit-data-manager/JaMMaTo/archive/refs/tags/3.0.2.tar.gz',    
  keywords = ['Schema', 'Mapping', 'DICOM', 'JSON'],   
  install_requires=[           
        'pydicom',
        'jsonschema',
        'urllib3',
        'datetime',
        'typing',
        'zipp',
        'nested_lookup'
      ],
  python_requires = '>=3.8',
  classifiers=[
    'Development Status :: 3 - Alpha',      
    'Intended Audience :: Developers',      
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: Common Public License',   
    'Programming Language :: Python :: 3'
  ],
)



import urllib.request
import logging
import json
import ssl
from .schemas_collector import schemas_collector_instance

class Cache_Schemas():

    def __init__(self, json_schema: dict) -> None:
        """Class instantiation.
        """
        self.json_schema = json_schema

    @classmethod
    def cache_schema(cls, map_dict: dict) -> dict:
        """Cache, or set the schema that is referenced by the map json document via the uri key and return it as dictionary.

        Args:
            map_dict (dict): _description_

        Raises:
            KeyError: No uri as key provided in the input dictionary.

        Returns:
            dict: The schema as dictionary.
        """
        ctx = ssl.create_default_context()
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE

        try:
            if schemas_collector_instance.get_uri(map_dict["uri"]):
                json_schema = schemas_collector_instance.get_schema(map_dict["uri"])
            else:
                with urllib.request.urlopen(map_dict["uri"], context=ctx) as url:
                    json_schema = json.load(url)
                schemas_collector_instance.add_schema(map_dict["uri"], json_schema)

        except KeyError as e:
            logging.error("Schema not accessible.")

        return cls(json_schema)